/*
 *  Blunder is an automated tool for analyzing chained exceptions.
 *  Copyright (C) 2009 - Ambrosi Lucas <mailto:lambrosi@users.sourceforge.net>
 *
 *  This file is part of Blunder.
 *
 *  Blunder is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License 
 *  version 3 as published by the Free Software Foundation.
 *
 *  Blunder is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public 
 *  License along with Blunder. If not, see <http://www.gnu.org/licenses/>.
 */

package net.sf.blunder;

import java.util.Collection;
import net.sf.blunder.model.ExceptionRelation;
import net.sf.blunder.util.BusinessExceptionDetector;
import ognl.Ognl;
import ognl.OgnlException;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;

/**
 * Blunder is an automated tool for analyzing chained exceptions in Java. 
 * 
 * @author <a href="mailto:lambrosi@users.sourceforge.net">Ambrosi Lucas</a>
 */
public class Blunder {

   private static final Logger log = Logger.getLogger( Blunder.class );

   @Autowired
   private BusinessExceptionDetector businessExceptionDetector;
   @Autowired
   private RelationResolver relationResolver;
   private ExceptionRelation exceptionRelation;
   private Throwable leafException;
   // WARN: next property could be null.
   private Throwable businessException;
   private Throwable rootException;

   /**
    * Default constructor.
    */
   public Blunder() {}

   /**
    * Process a {@link java.lang.Throwable} with Blunder.
    * @param t Chained Exception to analyze.
    */
   public Blunder process( Throwable t ) {
      log.debug( "Blunder initialized on: " + t.getMessage(), t );
      cleanUp();
      setLeafException( t );
      extractRootException();
      extractFirstBusinessException();
      lookupForExceptionRelation();
      return this;
   }

   /**
    * Cleans previous running information.
    */
   private void cleanUp() {
      log.debug( "Cleanning up previous running elements..." );
      setRootException( null );
      setBusinessException( null );
      setLeafException( null );
      setExceptionRelation( null );
   }

   /**
    * This method iterates the chained Exception
    * until the last Exception is found.
    */
   private void extractRootException() {
      Throwable cause = getLeafException();
      while ( cause.getCause() != null ) {
         cause = cause.getCause();
      }
      setRootException( cause );
   }

   /**
    * This method helps to find the first business 
    * Exception in the chain.
    */
   private void extractFirstBusinessException() {
      // Comparing not equals root and leaf and root not a business exception
      if ( getLeafException().getClass().equals( getRootException().getClass() ) ||
           isBusinessException( getRootException() ) ) {
         return;
      }
      Throwable e = getLeafException();
      Throwable firstBusiness = null;
      while ( e.getCause() != null ) {
         e = e.getCause();
         if ( isBusinessException( e ) ) {
            firstBusiness = e;
         }
      }
      setBusinessException( firstBusiness );
   }

   private void lookupForExceptionRelation() {
      setExceptionRelation( getRelationResolver().lookupRelation( this ) );
   }

   /**
    * @param t the Throwable code to analyze.
    * @return <b>true</b> if the {@link java.lang.Throwable} object is 
    * considered from business classes.
    *         <b>false</b> otherwise.
    */
   public boolean isBusinessException( Throwable t ) {
      return getBusinessExceptionDetector().isBusinessClass( t.getClass() );
   }

   /**
    * @return The highest level Exception in the chain.
    */
   public Throwable getLeafException() {
      return leafException;
   }

   /**
    * Sets the highest level Exception in the chain.
    */
   private void setLeafException( Throwable e ) {
      leafException = e;
   }

   /**
    * @return The business Exception in the chain. WARN:<b>Could be null.</b>
    */
   public Throwable getBusinessException() {
      return businessException;
   }

   private void setBusinessException( Throwable e ) {
      businessException = e;
   }

   /**
    * @return The lowest level Exception in the chain.
    */
   public Throwable getRootException() {
      return rootException;
   }

   private void setRootException( Throwable t ) {
      rootException = t;
   }

   /**
    * @return The {@link net.sf.blunder.model.ExceptionRelation} object.
    */
   public ExceptionRelation getExceptionRelation() {
      return exceptionRelation;
   }

   private void setExceptionRelation( ExceptionRelation er ) {
      exceptionRelation = er;
   }

   public BusinessExceptionDetector getBusinessExceptionDetector() {
      return businessExceptionDetector;
   }

   public void setBusinessExceptionDetector( BusinessExceptionDetector businessExceptionDetector ) {
      this.businessExceptionDetector = businessExceptionDetector;
   }

   public RelationResolver getRelationResolver() {
      return relationResolver;
   }

   public void setRelationResolver( RelationResolver relationResolver ) {
      this.relationResolver = relationResolver;
   }

   /**
    * @return the {@link java.lang.String} representation of
    * this chained Exception analysis.
    */
   @Override
   public String toString() {
      return new ToStringBuilder( this )
                  .append( "Leaf", getLeafException() )
                  .append( "Root", getRootException() )
                  .append( "Business", getBusinessException() )
                  .toString();
   }


   //////////////////////////////////////
   // Proyectors on ExceptionRelation
   //////////////////////////////////////

   /**
    * @return An array of {@link java.lang.String}s with the
    * list of possible solutions on this chained Exception.
    * @see net.sf.blunder.model.ExceptionRelation#getPossibleSolutionsList()
    */
   public String[] getPossibleSolutions() {
      try {
         return ( (Collection<?>)Ognl.getValue( getExceptionRelation().getPossibleSolutionsList(), this ) ).toArray( new String[]{} );
      } catch (OgnlException e) {
         log.warn( "Error parsing posible solutions list [ " + e.getMessage() +
                   "]. Replyin' the original expression [" + getExceptionRelation().getPossibleSolutionsList() + "]." );
         return new String[] { getExceptionRelation().getPossibleSolutionsList() };
      }
   }

   /**
    * @return A {@link java.lang.String} with the list of possible 
    * solutions on this chained Exception. Separated with a '\n'
    * character.
    * @see net.sf.blunder.model.ExceptionRelation#getPossibleSolutionsList()
    */
   public String getPossibleSolutionsAsString() {
      StringBuffer psols = new StringBuffer( "" );
      for ( String solution : getPossibleSolutions() ) {
         psols.append( solution );
         psols.append( "\n" );
      }
      return psols.toString();
   }

   /**
    * @return A {@link java.lang.String} with the list of possible 
    * solutions on this chained Exception. Represented as an HTML
    * list.
    * @see net.sf.blunder.model.ExceptionRelation#getPossibleSolutionsList()
    */
   public String getPossibleSolutionsAsHtmlList() {
      StringBuilder sb = new StringBuilder( 200 );
      sb.append( "<UL>\n" );
      for ( String solution: getPossibleSolutions() ) {
         sb.append( "<LI>" );
         sb.append( solution );
      }
      sb.append( "</UL>" );
      return sb.toString();
   }

   /**
    * @return The processed error message.
    * @see net.sf.blunder.model.ExceptionRelation#getFaultString()
    */
   public String getFaultString() {
      try {
         return (String) Ognl.getValue( getExceptionRelation().getFaultString(), this );
      } catch (OgnlException e) {
         log.warn( "Error parsing faultString [" + e.getMessage() +
                   "]. Replyin' the original expression [" + getExceptionRelation().getFaultString() + "]." );
         return getExceptionRelation().getFaultString();
      }
   }

   /**
    * @return The fault code {@link java.lang.String}.
    * @see net.sf.blunder.model.ExceptionRelation#getFaultCode()
    */
   public String getFaultCode() {
      return getExceptionRelation().getFaultCode();
   }

   public boolean isErrorKindOf( String kindOf ) {
      return getExceptionRelation().getFaultCode().equals( kindOf );
   }

   public boolean isServerError() {
      return isErrorKindOf( ExceptionRelation.SERVER_ERROR );
   }

   public boolean isBusinessError() {
      return isErrorKindOf( ExceptionRelation.BUSINESS_ERROR );
   }

   public boolean isClientError() {
      return isErrorKindOf( ExceptionRelation.CLIENT_ERROR );
   }

   public boolean isUserError() {
      return isErrorKindOf( ExceptionRelation.USER_ERROR );
   }

   public boolean isUndefinedError() {
      return isErrorKindOf( ExceptionRelation.UNDEFINED_ERROR );
   }

   public String getErrorDescription() {
      if ( isUserError() ) {
         return "User information error.";
      }
      if ( isClientError() ) {
         return "Client system error.";
      }
      if ( isBusinessError() ) {
         return "Business rules error.";
      }
      if ( isServerError() ) {
         return "System error.";
      }
      return "Undefined error cause.";
   }

}
