/*
 *  Blunder is an automated tool for analyzing chained exceptions.
 *  Copyright (C) 2009 - Ambrosi Lucas <mailto:lambrosi@users.sourceforge.net>
 *
 *  This file is part of Blunder.
 *
 *  Blunder is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License 
 *  version 3 as published by the Free Software Foundation.
 *
 *  Blunder is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public 
 *  License along with Blunder. If not, see <http://www.gnu.org/licenses/>.
 */

package net.sf.blunder;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.List;
import net.sf.blunder.model.ExceptionRelation;
import net.sf.blunder.service.persistence.BlunderPersistenceService;
import ognl.Ognl;
import ognl.OgnlException;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;

/**
 * This is where magic happens, this class is responsible for dealing
 * with errors.
 * 
 * <br/>
 * <i>This class is for internal use of Blunder.</i>
 * 
 * @author <a href="mailto:lambrosi@users.sourceforge.net">Ambrosi Lucas</a>
 */
public class RelationResolver {

   private static final Logger log = Logger.getLogger( RelationResolver.class );

   @Autowired
   private BlunderPersistenceService blunderPersistenceService;
   private ExceptionRelation defaultRelation;
   private Blunder blunder;
   private boolean shouldLearn;

   public RelationResolver() {}

   /**
    * Seeks the object representing the relationship for the
    * chained exception.
    * The search in the knowledge base is done through the following 
    * algorithm. At each step, if it doesn't find any results that 
    * match the current context error, continue with the next step.
    * 
    * <ol>
    *  <li>
    *    Tries to find a context with the three main elements, Leaf, Root and Business.
    *  </li>
    *  <li>
    *    If no match is found or doesn't meet the conditions, then try to find a
    *    context with the components, Leaf and Root.
    *  </li>
    *  <li>
    *    Then, it tries to find a context with the components Leaf and Business.
    *  </li>
    *  <li>
    *    Then tries to find a context with the compoent Leaf.
    *  </li>
    *  <li>
    *    Then tries to find a context with the compoent Root.
    *  </li>
    *  <li>
    *    And finally if none of those steps could solve the context, then it resolves
    *    with the default context. It is the one with the ID 0 (zero).
    *  </li>
    * </ol>
    * <p>
    *  Always is verified the expression in CONDITIONS and the availability of the
    *  context in runtime (MATCH_AT_RUNTIME='true').
    * </p>
    * <p>
    *  The main idea behind this series of steps is the generalization. Start with
    *  a search as specific as possible and go generalizing step by step. For example
    *  it is not necessary that all fields match the current situation; so we could
    *  take as rule that every chained exception that was originally generated by a
    *  {@link java.net.SocketException} is a comunication problem, therefore we would 
    *  like to address as a server error and suggest to the user to retry again in a
    *  few minutes. For this, we should specify a context with only this exception as
    *  root and all others learned contexts with {@link java.net.SocketException} as 
    *  Root should be marked as MATCH_AT_RUNTIME='false'.
    * </p>
    * 
    * @return The {@link net.sf.blunder.model.ExceptionRelation}
    * that matches the current chained exception.
    * @see net.sf.blunder.model.ExceptionRelation
    */
   public ExceptionRelation lookupRelation( Blunder blunder ) {
      this.blunder = blunder;
      try {
         log.debug( "Evaluating by Leaf, Root and FirstBusiness Exceptions..." );
         List<ExceptionRelation> candidates = getBlunderPersistenceService().findBy( getBlunder().getLeafException().getClass().getName(),
                                                                                     getBlunder().getRootException().getClass().getName(),
                                                                                     getBlunder().getBusinessException().getClass().getName() );
         return matchesRuntimeFilter( conditionsFilter( candidates ) );
      } catch ( Exception e1 ) {
         log.debug( "No valid relations found [" + e1.getMessage() + "]" );
         if ( getBlunder().getBusinessException() != null &&
              shouldLearn &&
              isAutolearning() ) {
            // Learns only when there is a business exception and its in autolearning
            // mode and there is a no-runtime relation in database.
            // In any other case it learns in other moment.
            return learnFromIt();
         }
         try {
            log.debug( "Evaluating by Leaf and Root Exceptions..." );
            List<ExceptionRelation> candidates = getBlunderPersistenceService().findByLeafAndRoot( getBlunder().getLeafException().getClass().getName(),
                                                                                                   getBlunder().getRootException().getClass().getName() );
            return matchesRuntimeFilter( conditionsFilter( candidates ) );
         } catch ( Exception e2 ) {
            log.debug( "No valid relations found [" + e2.getMessage() + "]" );
            if ( getBlunder().getBusinessException() == null &&
                 shouldLearn &&
                 isAutolearning() ) {
               // Learns only in autolearning mode autolearning, and there
               // is a no-runtime relation in database.
               // In any other case it learns in other moment.
               return learnFromIt();
            }
            try {
               log.debug( "Evaluating by Leaf and Business Exceptions..." );
               List<ExceptionRelation> candidates = getBlunderPersistenceService().findByLeafAndBusiness( getBlunder().getLeafException().getClass().getName(),
                                                                                                          getBlunder().getBusinessException().getClass().getName() );
               return matchesRuntimeFilter( conditionsFilter( candidates ) );
            } catch ( Exception e3 ) {
               log.debug( "No valid relations found [" + e3.getMessage() + "]" );
               try {
                  log.debug( "Evaluating by Leaf Exceptions..." );
                  List<ExceptionRelation> candidates = getBlunderPersistenceService().findByLeaf( getBlunder().getLeafException().getClass().getName() );
                  return matchesRuntimeFilter( conditionsFilter( candidates ) );
               } catch ( Exception e4 ) {
                  log.debug( "No valid relations found [" + e4.getMessage() + "]" );
                  try {
                     log.debug( "Evaluating by Root Exceptions..." );
                     List<ExceptionRelation> candidates = getBlunderPersistenceService().findByRoot( getBlunder().getRootException().getClass().getName() );
                     return matchesRuntimeFilter( conditionsFilter( candidates ) );
                  } catch ( Exception e5 ) {
                     // If we came up here, is beacause there is no relation in the database.
                     // We generate a Dummy relation to work on it.
                     log.warn( "No valid relations found [" + e5.getMessage() + "]. Generating a dummy." );
                     return getDummyExceptionRelation();
                  }
               }
            }
         }
      }
   }

   private List<ExceptionRelation> conditionsFilter( List<ExceptionRelation> candidates ) {
      log.debug( "Evaluating candidates..." );
      List<ExceptionRelation> conditionals = new ArrayList<ExceptionRelation>();
      for ( ExceptionRelation er : candidates ) {
         try {
            log.debug( "Evaluating relation [" + er +
                       "] with conditions [" + er.getConditions() +
                       "] on context [" + getBlunder() + "]" );
            if ( isDefault( er ) ) {
               defaultRelation = er;
            } else if ( ( (Boolean) Ognl.getValue( er.getConditions(), getBlunder() ) ).booleanValue() ) {
               // It follows the conditions, its a candidate
               conditionals.add( er );
            }
         } catch ( OgnlException oe ) {
            // It can be an error on the expression or maybe it is not
            // evaluating in an expected context.
            log.warn( "Error evaluating expression [" + er.getConditions() +
                      "] for the relation [" + er +
                      "] on context [" + getBlunder() +
                      "]. Caused by [" + oe.getMessage() + "]" );
         }
      }
      return conditionals;
   }

   /**
    * @return <b>true</b> if this is the default relation for
    * the context.
    */
   private boolean isDefault( ExceptionRelation er ) {
      return er.getConditions() == null ||
             "".equals( er.getConditions() ) ||
             er.getConditions().trim().equalsIgnoreCase( "true" );
   }

   private ExceptionRelation matchesRuntimeFilter( List<ExceptionRelation> candidates ) throws LookupException {
      // If we are here its beacause we have strong candidates, just left to know
      // if any of them is prepared to solve at runtime. At this point we should
      // not learn.
      shouldLearn = candidates.isEmpty();
      for ( ExceptionRelation er : candidates ) {
         // If it is enabled for runtime, is just what we found!
         if ( er.isRuntimeMatch() ) {
            log.debug( "Runtime found for relation [" + er + "]" );
            return er;
         }
      }
      // WARN: we asume that the default relation, if exists, always
      // are available for runtime.
      if ( defaultRelation != null ) {
         log.debug( "Runtime match by default found for relation [" + defaultRelation + "]" );
         return defaultRelation;
      }
      // None relation is enabled for runtime and there is no default relation.
      throw new LookupException( "No candidate is enable for runtime." );
   }

   private ExceptionRelation getDummyExceptionRelation() {
      ExceptionRelation er = null;
      Throwable leaf = getBlunder().getLeafException();
      log.warn( "Solving on dummy relation [" + leaf.getMessage() + "]. Reference stacktrace: ", leaf );
      try {
         er = getBlunderPersistenceService().findDefaultRelation().clone();
      } catch ( Exception e ) {
         log.warn( "Couldn't solve default ExceptionRelation from database. Generating one locally.", e );
         er = new ExceptionRelation();
         er.setFaultCode( ExceptionRelation.UNDEFINED_ERROR );
         // FaultString is an OGNL expresion.
         er.setFaultString( "\"Undefined error cause.\"" );
         er.setRuntimeMatch( true );
         // Posible solutions list in OGNL.
         er.setPossibleSolutionsList( "{ \"Retry in few minutes, if the error persist call help desk.\" }" );
      }
      er.setLeafExceptionClassname( getBlunder().getLeafException().getClass().getName() );
      er.setRootExceptionClassname( getBlunder().getRootException().getClass().getName() );
      if ( getBlunder().getBusinessException() != null ) {
         er.setBusinessExceptionClassname( getBlunder().getBusinessException().getClass().getName() );
      }
      return er;
   }

   private ExceptionRelation learnFromIt() {
      log.info( "Trying to learn [" + getBlunder() + "]" );

      // We take the dummy relation and change some stuff.
      ExceptionRelation er = getDummyExceptionRelation();
      StringWriter sw = new StringWriter();
      getBlunder().getLeafException().printStackTrace( new PrintWriter( sw ) );
      sw.flush();
      String stack = sw.toString();
      // WARN: VARCHAR2 database limit!!!
      if ( stack.length() >= 4000 ) {
         stack = stack.substring( 0, 3999 );
      }
      er.setDebugStackTrace( stack );
      try { sw.close(); } catch (IOException e1) {}

      if ( isPureJavaLangExceptionContext() ) {
         log.warn( "We are learning a pure java.lang.Exception classes!!" );
         StackTraceElement ste = getBlunder().getLeafException().getStackTrace()[0];
         er.setConditions( "( leafException.stackTrace[0].className.equals(  \"" + ste.getClassName()+ "\") && " +
                           "  leafException.stackTrace[0].methodName.equals( \"" + ste.getMethodName() + "\" ) )" );
      }

      try {
         getBlunderPersistenceService().learn( er );
         log.info( "We have learn a new error [" + er + "], next time we'll get it better! :D" );
      } catch (Exception e) {
         // These method only runs in develop mode, that's why we print the full stacktrace!
         log.error( "Can't learn relation [" + er + "] beacause of [" + e.getMessage() + "]", e );
      }
      return er;
   }

   private boolean isPureJavaLangExceptionContext() {
      return getBlunder().getLeafException().getClass().getName().equals( Exception.class.getName() ) &&
             getBlunder().getRootException().getClass().getName().equals( Exception.class.getName() ) &&
             getBlunder().getBusinessException() == null;
   }

   private boolean isAutolearning() {
      return System.getProperty( "net.sf.blunder.autolearning", "false" ).trim().equalsIgnoreCase( "true" );
   }

   private Blunder getBlunder() {
      return blunder;
   }

   public BlunderPersistenceService getBlunderPersistenceService() {
      return blunderPersistenceService;
   }

   public void setBlunderPersistenceService( BlunderPersistenceService blunderPersistenceService ) {
      this.blunderPersistenceService = blunderPersistenceService;
   }

}
