/*
 *  Blunder is an automated tool for analyzing chained exceptions.
 *  Copyright (C) 2009 - Ambrosi Lucas <mailto:lambrosi@users.sourceforge.net>
 *
 *  This file is part of Blunder.
 *
 *  Blunder is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License 
 *  version 3 as published by the Free Software Foundation.
 *
 *  Blunder is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public 
 *  License along with Blunder. If not, see <http://www.gnu.org/licenses/>.
 */

package net.sf.blunder.model;

import java.io.Serializable;
import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;
import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * The model for all the data of an error context.
 * 
 * <br/>
 * <i>This class is for internal use of Blunder.</i>
 * 
 * @author <a href="mailto:lambrosi@users.sourceforge.net">Ambrosi Lucas</a>
 */
public class ExceptionRelation implements Cloneable, Serializable {

   public static final long serialVersionUID = 20080621182759000L;

   /** Undefined error type. */
   public static String UNDEFINED_ERROR   = "Undefined";
   /** A Server error type. */
   public static String SERVER_ERROR      = "Server";
   /** A Business error type. */
   public static String BUSINESS_ERROR    = "Business";
   /** A Client error type. */
   public static String CLIENT_ERROR      = "Client";
   /** A User error type. */
   public static String USER_ERROR        = "User";

   private Long id;
   private String conditions;
   private String businessExceptionClassname;
   private String leafExceptionClassname;
   private String rootExceptionClassname;
   private String possibleSolutionsList;
   private String faultString;
   private String faultCode;
   /**
    * The relation is in the database, so there is no need to learn it
    * but it means 'do not solve by this relation at runtime'. Probably
    * there is another relation more generic in the database.
    */
   private boolean matchAtRuntime;
   private String debugStackTrace;

   /**
    * @return The ExceptionRelation Id.
    */
   public Long getId() {
      return id;
   }

   /**
    * Sets the ExceptionRelation Id.
    */
   public void setId( Long l ) {
      id = l;
   }

   /**
    * @return the conditions expression that any context should 
    * apply for resolving with this particular ExceptionRelation.
    */
   public String getConditions() {
      return conditions;
   }

   /**
    * Sets the conditions expression that any context should 
    * apply for resolving with this particular ExceptionRelation.
    */
   public void setConditions( String s ) {
      conditions = s;
   }

   /**
    * @return the fault code associated with this ExceptionRelation.
    */
   public String getFaultCode() {
      return faultCode;
   }

   /**
    * Sets the fault code associated with this ExceptionRelation.
    */
   public void setFaultCode( String s ) {
      faultCode = s;
   }

   /**
    * @return the fault string expression associated with this
    * ExceptionRelation.
    */
   public String getFaultString() {
      return faultString;
   }

   /**
    * Sets the fault string expression associated with this
    * ExceptionRelation.
    */
   public void setFaultString( String s ) {
      faultString = s;
   }

   /**
    * @return the Business exception class name associated with this
    * ExceptionRelation.
    */
   public String getBusinessExceptionClassname() {
      return businessExceptionClassname;
   }

   /**
    * Sets the Business exception class name associated with this
    * ExceptionRelation.
    */
   public void setBusinessExceptionClassname( String s ) {
      businessExceptionClassname = s;
   }

   /**
    * @return the Leaf exception class name associated with this
    * ExceptionRelation.
    */
   public String getLeafExceptionClassname() {
      return leafExceptionClassname;
   }

   /**
    * Sets the Leaf exception class name associated with this
    * ExceptionRelation.
    */
   public void setLeafExceptionClassname( String s ) {
      leafExceptionClassname = s;
   }

   /**
    * @return the possible solutions list expression associated
    * with this ExceptionRelation.
    */
   public String getPossibleSolutionsList() {
      return possibleSolutionsList;
   }

   /**
    * Sets the possible solutions list expression associated
    * with this ExceptionRelation.
    */
   public void setPossibleSolutionsList( String s ) {
      possibleSolutionsList = s;
   }

   /**
    * @return the Root exception class name associated with this
    * ExceptionRelation.
    */
   public String getRootExceptionClassname() {
      return rootExceptionClassname;
   }

   /**
    * Sets the Root exception class name associated with this
    * ExceptionRelation.
    */
   public void setRootExceptionClassname( String s ) {
      rootExceptionClassname = s;
   }

   /**
    * Just for mapping uniformely across diferent dbms.
    * @see <code>ExceptionRelation.isRuntimeMatch()</code>
    */
   @Deprecated
   public String getMatchAtRuntime() {
      return isRuntimeMatch() ? "Y" : "N" ;
   }

   /**
    * Just for mapping uniformely across diferent dbms.
    * @see <code>#setRuntimeMatch( boolean )</code>
    */
   @Deprecated
   public void setMatchAtRuntime( String s ) {
      matchAtRuntime = s.equalsIgnoreCase( "Y" );
   }

   /**
    * Sets if this ExceptionRelation should match at runtime.
    */
   public void setRuntimeMatch( boolean b ) {
      matchAtRuntime = b;
   }

   /**
    * @return <b>true</b> if this ExceptionRelation should match
    * at runtime.
    */
   public boolean isRuntimeMatch() {
      return matchAtRuntime;
   }

   /**
    * @return the debug stacktrace for this ExceptionRelation.
    */
   public String getDebugStackTrace() {
      return debugStackTrace;
   }

   /**
    * Sets the debug stacktrace for this ExceptionRelation.
    */
   public void setDebugStackTrace( String s ) {
      debugStackTrace = s;
   }

   @Override
   public String toString() {
      return new ToStringBuilder( this )
                  .append( "Id", getId() )
                  .append( "Leaf", getLeafExceptionClassname() )
                  .append( "Root", getRootExceptionClassname() )
                  .append( "Business", getBusinessExceptionClassname() )
                  .append( "FaultCode", getFaultCode() )
                  .append( "FaultString", getFaultString() )
                  .toString();
   }

   @Override
   public int hashCode() {
      return new HashCodeBuilder()
                  .append( getLeafExceptionClassname() )
                  .append( getRootExceptionClassname() )
                  .append( getBusinessExceptionClassname() )
                  .append( getConditions() )
                  .toHashCode();
   }

   @Override
   public boolean equals(Object other) {
      if ( !(other instanceof ExceptionRelation ) ) return false;
      return new EqualsBuilder()
                  .append( this.hashCode(), other.hashCode() )
                  .isEquals();
   }

   @Override
   public ExceptionRelation clone() throws CloneNotSupportedException {
      ExceptionRelation clone = (ExceptionRelation) super.clone();
      clone.setId( null );
      return clone;
   }

}
