/*
 *  Blunder is an automated tool for analyzing chained exceptions.
 *  Copyright (C) 2009 - Ambrosi Lucas <mailto:lambrosi@users.sourceforge.net>
 *
 *  This file is part of Blunder.
 *
 *  Blunder is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License 
 *  version 3 as published by the Free Software Foundation.
 *
 *  Blunder is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public 
 *  License along with Blunder. If not, see <http://www.gnu.org/licenses/>.
 */

package net.sf.blunder.service.persistence;

import java.util.List;
import net.sf.blunder.model.ExceptionRelation;

/**
 * Interface for defining a persistence service for Blunder.
 * If you plan to write your own persistence mechanism, you
 * should start writing your own implementation of this class.
 * 
 * @author <a href="mailto:lambrosi@users.sourceforge.net">Ambrosi Lucas</a>
 * @see net.sf.blunder.service.persistence.InMemoryPersistenceService
 */
public interface BlunderPersistenceService {

   /**
    * Finds an {@link net.sf.blunder.model.ExceptionRelation} for
    * the following arguments.
    *  
    * @param leafClassname class name of the highest level
    *        {@link java.lang.Exception} in the chain.
    * @param rootClassname class name of the lowest level
    *        {@link java.lang.Exception} in the chain.
    * @param businessClassname class name of the business 
    *        {@link java.lang.Exception} in the chain.
    * @return a {@link java.util.List} of {@link net.sf.blunder.model.ExceptionRelation}
    *        with all relations that matches with the previous parameters.
    * @see net.sf.blunder.RelationResolver
    */
   public List<ExceptionRelation> findBy( String leafClassname, String rootClassname, String businessClassname );

   /**
    * Finds an {@link net.sf.blunder.model.ExceptionRelation} for
    * the following arguments.
    *  
    * @param leafClassname class name of the highest level
    *        {@link java.lang.Exception} in the chain.
    * @param businessClassname class name of the business 
    *        {@link java.lang.Exception} in the chain.
    * @return a {@link java.util.List} of {@link net.sf.blunder.model.ExceptionRelation}
    *        with all relations that matches with the previous parameters.
    * @see net.sf.blunder.RelationResolver
    */
   public List<ExceptionRelation> findByLeafAndBusiness( String leafClassname, String businessClassname );

   /**
    * Finds an {@link net.sf.blunder.model.ExceptionRelation} for
    * the following arguments.
    *  
    * @param leafClassname class name of the highest level
    *        {@link java.lang.Exception} in the chain.
    * @param rootClassname class name of the lowest level
    *        {@link java.lang.Exception} in the chain.
    * @return a {@link java.util.List} of {@link net.sf.blunder.model.ExceptionRelation}
    *        with all relations that matches with the previous parameters.
    * @see net.sf.blunder.RelationResolver
    */
   public List<ExceptionRelation> findByLeafAndRoot( String leafClassname, String rootClassname );

   /**
    * Finds an {@link net.sf.blunder.model.ExceptionRelation} for
    * the following arguments.
    *  
    * @param rootClassname class name of the lowest level
    *        {@link java.lang.Exception} in the chain.
    * @return a {@link java.util.List} of {@link net.sf.blunder.model.ExceptionRelation}
    *        with all relations that matches with the previous parameters.
    * @see net.sf.blunder.RelationResolver
    */
   public List<ExceptionRelation> findByRoot( String rootClassname );

   /**
    * Finds an {@link net.sf.blunder.model.ExceptionRelation} for
    * the following arguments.
    *  
    * @param leafClassname class name of the highest level
    *        {@link java.lang.Exception} in the chain.
    * @return a {@link java.util.List} of {@link net.sf.blunder.model.ExceptionRelation}
    *        with all relations that matches with the previous parameters.
    * @see net.sf.blunder.RelationResolver
    */
   public List<ExceptionRelation> findByLeaf( String leafClassname );

   /**
    * Finds an {@link net.sf.blunder.model.ExceptionRelation} for
    * the following arguments.
    *  
    * @param businessClassname class name of the business 
    *        {@link java.lang.Exception} in the chain.
    * @return a {@link java.util.List} of {@link net.sf.blunder.model.ExceptionRelation}
    *        with all relations that matches with the previous parameters.
    * @see net.sf.blunder.RelationResolver
    */
   public List<ExceptionRelation> findByBusiness( String businessClassname );

   /**
    * @return the default {@link net.sf.blunder.model.ExceptionRelation}.
    */
   public ExceptionRelation findDefaultRelation();

   /**
    * Learns a new {@link net.sf.blunder.model.ExceptionRelation}.
    * @param relation the {@link net.sf.blunder.model.ExceptionRelation} to learn.
    */
   public void learn( ExceptionRelation relation );

   /**
    * Updates a known {@link net.sf.blunder.model.ExceptionRelation}.
    * @param relation the {@link net.sf.blunder.model.ExceptionRelation} to update.
    */
   public void update( ExceptionRelation relation );

   /**
    * Deletes a new {@link net.sf.blunder.model.ExceptionRelation}.
    * @param relation the {@link net.sf.blunder.model.ExceptionRelation} to delete.
    */
   public void delete( ExceptionRelation relation );

   /**
    * Refresh the persistence service. This should be used to clear all
    * data in the local cache.
    */
   public void refresh();

}
