/*
 *  Blunder is an automated tool for analyzing chained exceptions.
 *  Copyright (C) 2009 - Ambrosi Lucas <mailto:lambrosi@users.sourceforge.net>
 *
 *  This file is part of Blunder.
 *
 *  Blunder is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License 
 *  version 3 as published by the Free Software Foundation.
 *
 *  Blunder is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public 
 *  License along with Blunder. If not, see <http://www.gnu.org/licenses/>.
 */

package net.sf.blunder.service.persistence.binary;

import java.io.EOFException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.HashMap;
import net.sf.blunder.BlunderException;
import net.sf.blunder.model.ExceptionRelation;
import net.sf.blunder.service.persistence.InMemoryPersistenceService;
import org.apache.log4j.Logger;

/**
 * Binary implementation for a {@link net.sf.blunder.service.persistence.BlunderPersistenceService}.
 * It reads and writes data from/to a Java serialized file.
 * 
 * @author <a href="mailto:lambrosi@users.sourceforge.net">Ambrosi Lucas</a>
 */
public class BlunderBinaryService extends InMemoryPersistenceService {

   private static final Logger log = Logger.getLogger( BlunderBinaryService.class );

   /**
    * Returns the value as it's passed.
    * @param value
    * @return the value as it's passed.
    * @see net.sf.blunder.service.persistence.InMemoryPersistenceService#getEmptyOrValueString()
    */
   @Override
   protected String getEmptyOrValueString( String value ) {
      return value;
   }

   /**
    * Binary implementation of the data reader.
    * @see net.sf.blunder.service.persistence.InMemoryPersistenceService#readData()
    */
   @Override
   public HashMap<Long, ExceptionRelation> readData() {
      ObjectInputStream ois = null;
      HashMap<Long, ExceptionRelation> map = new HashMap<Long, ExceptionRelation>();
      try {
         log.debug( "Reading data from file [" + getFilePath() + "]" );
         ois = new ObjectInputStream( getInputStream() );
         ExceptionRelation obj = null;
         while( ( obj = (ExceptionRelation)ois.readObject() ) != null ) {
            if ( log.isDebugEnabled() ) log.debug( "Adding to map [" + obj + "] from file!" );
            map.put( obj.getId(), obj );
         }
         return map;
      } catch ( EOFException eof ) {
         // Search for a better way for detect EOF!!!
         log.debug( "EOF reached!" );
         return map;
      } catch ( Exception e ) {
         throw new BlunderException( "Couldn't read object due to " + e.getMessage(), e );
      } finally {
         if ( ois != null ) {
            try { ois.close(); } catch ( Exception e ) {}
         }
      }
   }

   /**
    * Binary implementation of the data writer.
    * @see net.sf.blunder.service.persistence.InMemoryPersistenceService#writeData()
    */
   @Override
   public void writeData() {
      ObjectOutputStream oos = null;
      try {
         oos = new ObjectOutputStream( getOutputStream() );
         for ( ExceptionRelation er : getData().values() ) {
            if ( log.isDebugEnabled() ) log.debug( "Writing [" + er + "]" );
            oos.writeObject( er );
            if ( log.isDebugEnabled() ) log.debug( "Done [" + er + "]" );
         }
      } catch ( Exception e ) {
         throw new BlunderException( "Couldn't write object due to " + e.getMessage(), e );
      } finally {
         if ( oos != null ) {
            try { oos.flush(); } catch ( Exception e ) {}
            try { oos.close(); } catch ( Exception e ) {}
         }
      }
   }

}
