/*
 *  Blunder is an automated tool for analyzing chained exceptions.
 *  Copyright (C) 2009 - Ambrosi Lucas <mailto:lambrosi@users.sourceforge.net>
 *
 *  This file is part of Blunder.
 *
 *  Blunder is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License 
 *  version 3 as published by the Free Software Foundation.
 *
 *  Blunder is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public 
 *  License along with Blunder. If not, see <http://www.gnu.org/licenses/>.
 */

package net.sf.blunder.service.persistence.hibernate;

import java.util.List;
import javax.annotation.Resource;
import net.sf.blunder.model.ExceptionRelation;
import net.sf.blunder.service.persistence.BlunderPersistenceService;
import org.apache.log4j.Logger;
import org.hibernate.Criteria;
import org.hibernate.Session;
import org.hibernate.SessionException;
import org.hibernate.criterion.Criterion;
import org.hibernate.criterion.Restrictions;
import org.springframework.dao.DataAccessException;
import org.springframework.orm.hibernate3.HibernateTemplate;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

/**
 * Hibernate implementation for a {@link net.sf.blunder.service.persistence.BlunderPersistenceService}.
 * For more information about Hibernate please refer to the project site
 * <a href="http://www.hibernate.org">http://www.hibernate.org</a>.
 * 
 * @author <a href="mailto:lambrosi@users.sourceforge.net">Ambrosi Lucas</a>
 */
@Transactional(readOnly = true, propagation = Propagation.NOT_SUPPORTED)
public class BlunderHibernateService implements BlunderPersistenceService {

   private static final Logger log = Logger.getLogger( BlunderHibernateService.class );

   @Resource
   private HibernateTemplate hibernateTemplate;

   /**
    * @see net.sf.blunder.service.persistence.BlunderPersistenceService#findBy(String,String,String)
    */
   @SuppressWarnings("unchecked")
   public List<ExceptionRelation> findBy( String leafClassname, String rootClassname, String businessClassname ) throws DataAccessException, SessionException {
      Session s = getHibernateTemplate().getSessionFactory().openSession();
      try {
         Criteria criteria = s.createCriteria( ExceptionRelation.class );
         Criterion leafExpression = createExpression( "leafExceptionClassname", leafClassname );
         Criterion rootExpression = createExpression( "rootExceptionClassname", rootClassname );
         Criterion businessExpression = createExpression( "businessExceptionClassname", businessClassname );

         log.debug( "About to execute query..." );
         return criteria.add( Restrictions.and( leafExpression,
                                                Restrictions.and( rootExpression,
                                                                  businessExpression ) ) )
                        .list();
      } catch ( Exception e ) {
         log.warn( "Error running query: " + e.getMessage(), e );
         return null;
      } finally {
         if ( s != null ) {
            s.close();
         }
      }
   }

   /**
    * Creates an expression for Hibernate Criteria.
    * @param property name
    * @param property value
    * @return the Criterion for the expression.
    */
   private Criterion createExpression( String propertyName, String value ) {
      if ( value == null ) {
         log.debug( "Not a value for property [" + propertyName + "] query for excluding!" );
         return Restrictions.or( Restrictions.isNull( propertyName ),
                                 Restrictions.eq( propertyName, "" ) );
      }
      log.debug( "Setting value [" + value + "] for property [" + propertyName + "]" );
      return Restrictions.eq( propertyName, value );
   }

   /**
    * @see net.sf.blunder.service.persistence.BlunderPersistenceService#findByBusiness(String)
    */
   public List<ExceptionRelation> findByBusiness( String businessClassname ) {
      return findBy( null, null, businessClassname );
   }

   /**
    * @see net.sf.blunder.service.persistence.BlunderPersistenceService#findByLeaf(String)
    */
   public List<ExceptionRelation> findByLeaf( String leafClassname ) {
      return findBy( leafClassname, null, null );
   }

   /**
    * @see net.sf.blunder.service.persistence.BlunderPersistenceService#findByLeafAndBusiness(String,String)
    */
   public List<ExceptionRelation> findByLeafAndBusiness( String leafClassname, String businessClassname ) {
      return findBy( leafClassname, null, businessClassname );
   }

   /**
    * @see net.sf.blunder.service.persistence.BlunderPersistenceService#findByLeafAndRoot(String,String)
    */
   public List<ExceptionRelation> findByLeafAndRoot( String leafClassname, String rootClassname ) {
      return findBy( leafClassname, rootClassname, null );
   }

   /**
    * @see net.sf.blunder.service.persistence.BlunderPersistenceService#findByRoot(String)
    */
   public List<ExceptionRelation> findByRoot( String rootClassname ) {
      return findBy( null, rootClassname, null );
   }

   /**
    * @see net.sf.blunder.service.persistence.BlunderPersistenceService#findDefaultRelation()
    */
   public ExceptionRelation findDefaultRelation() {
      log.debug( "Lookup for Default ExceptionRelation..." );
      return (ExceptionRelation) getHibernateTemplate().get( ExceptionRelation.class, new Long( 0 ) );
   }

   /**
    * @see net.sf.blunder.service.persistence.BlunderPersistenceService#learn(ExceptionRelation)
    */
   @Transactional(readOnly = false, propagation = Propagation.SUPPORTS)
   public void learn( ExceptionRelation er ) throws DataAccessException, SessionException {
      log.debug( "Persisting relation [" + er + "]" );
      getHibernateTemplate().persist( er );
   }

   /**
    * @see net.sf.blunder.service.persistence.BlunderPersistenceService#update(ExceptionRelation)
    */
   @Transactional(readOnly = false, propagation = Propagation.SUPPORTS)
   public void update( ExceptionRelation er ) {
      log.debug( "Updating relation [" + er + "]" );
      getHibernateTemplate().update( er );
   }

   /**
    * @see net.sf.blunder.service.persistence.BlunderPersistenceService#delete(ExceptionRelation)
    */
   @Transactional(readOnly = false, propagation = Propagation.SUPPORTS)
   public void delete( ExceptionRelation er ) {
      log.debug( "Deleting relation [" + er + "]" );
      getHibernateTemplate().delete( er );
   }

   /**
    * @see net.sf.blunder.service.persistence.BlunderPersistenceService#refresh()
    */
   public void refresh() {
      log.debug( "Cache clear..." );
      getHibernateTemplate().clear();
   }

   public HibernateTemplate getHibernateTemplate() {
      return hibernateTemplate;
   }

   public void setHibernateTemplate( HibernateTemplate hibernateTemplate ) {
      this.hibernateTemplate = hibernateTemplate;
   }

}
